/*
  TimeSheet
    あかつきみさき

  このスクリプトについて
    選択したレイヤーにデジタルタイムシートファイル(.tdts/.xdts)から参照した動画のタイムリマップデータを適用します.
      (現在のバージョンでは原画には対応していません.原画対応の場合は中割り記号などをなくして,数字を正の整数で記述することで適用できます)

    タイムリマップはセルと同じ名前のデータが適用されます.
    ダブラシなど,アンダーバー(_)で繋がったレイヤーがある時は同一名のシートがない時はアンダーバーを取り除いたシートが適用されます.
    例: a_wxpレイヤーがあり,シートデータにa_wxpがない時,aのシートデータが適用

    .tdtsファイルは「東映アニメーション デジタルタイムシート for Windows」のファイルです.
    .xdtsファイルは「CLIP STUDIO PAINT」 (Ver.1.8.4以降)、「東映アニメーション デジタルタイムシート for Windows」から書き出すことができます.

  参考リンク
    東映アニメーション デジタルタイムシート for Windows
      https://www.clip-studio.com/clip_site/download/clipstudiopaint/toeianimation/index

  CLIP STUDIO PAINT
      https://www.clip-studio.com/clip_site/mypage/newsdetail?publish=201811&f=2018112901

  サポートしているデジタルタイムシートのバージョン
      .tdts (東映アニメーションデジタルタイムシート)
        version 5

      .xdts (汎用デジタルタイムシート)
        version 5

  使用方法
    コンポジションをアクティブにしてスクリプトを実行.
    レイヤーを選択してスクリプトを実行.

  バージョン情報
    2019/04/27 Ver 1.1.0 Update
        シートの最大値がセルの枚数より大きい時,不正なタイムリマップが適用される問題の修正.
        シートの最大値とセルの枚数が一致しない時,警告を出すようにした.

    2019/02/01 Ver 1.0.1 Update
        フィールドにvaluesが存在しない時,エラーが出る問題の修正.
        最後が空セルの時,ブラインドの最後のキーフレームが適用されない問題の修正.
        ブラインドの変換終了の最初のキーフレーム値が,シートに関係なく100になる問題の修正.
        タイムリマップの最初のキーフレーム値が,シートに関係なく1になる問題の修正.

    2018/12/24 Ver 1.0.0 Release
        シート尺適用機能の追加.
        ボールド自動挿入機能の追加.
        同セル名が存在せず,アンダーバー区切りで文字列が一致するシートが存在する時,同セルとして処理するようにした.
        レイヤーを選択せずに実行した時,すべてのレイヤーを調べ,レイヤー名と一致するものに適用するようにした.
        ブラインドエフェクトが存在する時,ブラインドを削除し再適用するようにした.
        UIの追加.
        CS6への対応.

    2018/11/30 Ver 0.2.0 Update
        .tdtsに対応.
        先頭が空セルの時,ブラインドが正しく入力されない問題の修正.
        セリフやカメラワークのフィールドが存在する時,セルに異なるデータが入力される問題の修正.

    2018/11/29 Ver 0.1.1 Update
        セルのアルファベットを大文字小文字で区別しないようにした.

    2018/11/29 Ver 0.1.0 PreRelease
*/

/// <reference path="D:\lib\aftereffects.d.ts\ae.d.ts"/>

// @include "json2.js"

(function () {
    Settings.prototype.initSetting = function (sectionName, keyName, defauleValue) {
        if (!app.settings.haveSetting(sectionName, keyName)) {
            app.settings.saveSetting(sectionName, keyName, defauleValue);
        }

        return app.settings.getSetting(sectionName, keyName);
    }

    ItemCollection.prototype.generateBoardComp = function (name, width, height, comp, textColor, bgColor) {
        var FRAME_COLOR_DEFAULT = [0, 0, 0];
        var BG_COLOR_DEFAULT = [1, 1, 1];

        var boardComp = this.addComp(name, width, height, comp.pixelAspect, 10800, comp.frameRate);

        var shapeLayer = boardComp.layers.addShape();
        shapeLayer.name = "frame";
        var shapeRoot = shapeLayer.property("ADBE Root Vectors Group");
        var shapePath = shapeRoot.addProperty("ADBE Vector Shape - Group");
        var shape = new Shape()
        shape.vertices = [[0, 0], [width, 0], [width, height], [0, height]];
        shape.closed = true;
        shapePath.property("ADBE Vector Shape").setValue(shape);

        shapeRoot.addProperty("ADBE Vector Graphic - Stroke");
        shapeRoot.addProperty("ADBE Vector Graphic - Fill");

        var shapeStroke = shapeRoot.property("ADBE Vector Graphic - Stroke");
        var shapeFill = shapeRoot.property("ADBE Vector Graphic - Fill");

        shapeStroke.property("ADBE Vector Stroke Color").setValue(FRAME_COLOR_DEFAULT);
        shapeStroke.property("ADBE Vector Stroke Width").setValue(Math.round(height / 20));
        shapeFill.property("ADBE Vector Fill Color").setValue(BG_COLOR_DEFAULT);

        var rect = shapeLayer.sourceRectAtTime(boardComp.time, true);
        shapeLayer.property("ADBE Transform Group").property("ADBE Anchor Point").setValue([Math.round(rect.left + rect.width / 2), Math.round(rect.top + rect.height / 2)]);
        shapeLayer.property("ADBE Transform Group").property("ADBE Position").setValue([Math.round(boardComp.width / 2), Math.round(boardComp.height / 2)]);

        var textLayer = boardComp.layers.addText(new TextDocument(name));
        textLayer.name = "text";
        var rect = textLayer.sourceRectAtTime(boardComp.time, true);
        textLayer.property("ADBE Transform Group").property("ADBE Anchor Point").setValue([Math.round(rect.left + rect.width / 2), Math.round(rect.top + rect.height / 2)]);
        textLayer.property("ADBE Transform Group").property("ADBE Position").setValue([Math.round(boardComp.width / 2), Math.round(boardComp.height / 2)]);

        var margin = boardComp.width / 6;
        textLayer.property("ADBE Transform Group").property("ADBE Scale").setValue([Math.round((boardComp.width - margin) / rect.width * 100), Math.round((boardComp.height - margin) / rect.height * 100)]);

        var eParade = textLayer.property("ADBE Effect Parade");

        // 文字
        var textColorEx = eParade.addProperty("ADBE Color Control");
        textColorEx.name = "textColor";
        textColorEx.property("ADBE Color Control-0001").setValue(textColor);

        var bgColorEx = eParade.addProperty("ADBE Color Control");
        bgColorEx.name = "bgColor";
        bgColorEx.property("ADBE Color Control-0001").setValue(bgColor);

        eParade.addProperty("ADBE Fill");
        eParade.property("ADBE Fill").property("ADBE Fill-0002").setValue(FRAME_COLOR_DEFAULT);
        eParade.property("ADBE Fill").property("ADBE Fill-0002").expression = 'thisComp.layer("text").effect("textColor")("ADBE Color Control-0001")';

        shapeStroke.property("ADBE Vector Stroke Color").expression = 'thisComp.layer("text").effect("textColor")("ADBE Color Control-0001")';
        shapeFill.property("ADBE Vector Fill Color").expression = 'thisComp.layer("text").effect("bgColor")("ADBE Color Control-0001")';
        return boardComp;
    }

    // 同じファイルがプロジェクトに存在する時は読みこまない
    Project.prototype.importFileSmart = function (inportOpt) {
        for (var i = 1; i <= app.project.numItems; i++) {
            if (!(app.project.item(i) instanceof FootageItem)) {
                continue;
            }

            if (app.project.item(i).file == inportOpt.file) {
                return app.project.item(i);
            }
        }

        return app.project.importFile(inportOpt);
    }

    Project.prototype.getItemsFromComment = function (comment) {
        var sameNameItems = [];
        for (var i = 1; i <= app.project.numItems; i++) {
            if (app.project.item(i).comment == comment) {
                sameNameItems[sameNameItems.length] = app.project.item(i);
            }
        }

        return sameNameItems;
    }

    Project.prototype.getItemsFromFile = function (file) {
        var sameFileItems = [];
        for (var i = 1; i <= app.project.numItems; i++) {
            if (app.project.item(i) instanceof FootageItem) {
                continue;
            }

            if (app.project.item(i).file == file) {
                sameFileItems[sameFileItems.length] = app.project.item(i);
            }
        }

        return sameFileItems;
    }

    Folder.prototype.getBoardImage = function (fileName) {
        var boardImages = this.getFiles(/.jpg$|.jpeg$|.png$|.tga$/);

        for (var i = 0; i < boardImages.length; i++) {
            if (fileName == decodeURIComponent(boardImages[i].name.split(".")[0])) {
                return boardImages[i];
            }
        }
        return null;
    }


    var SECTION_NAME = "TimeSheet";
    var APPLY_SHEET = Boolean(Number(app.settings.initSetting(SECTION_NAME, "APPLY_SHEET", 1)) == 1);
    var ADD_BOARD = Boolean(Number(app.settings.initSetting(SECTION_NAME, "ADD_BOARD", 1)) == 1);
    var APPLY_DURATION = Boolean(Number(app.settings.initSetting(SECTION_NAME, "APPLY_DURATION", 1)) == 1);

    var WIDTH_SIZE = Number(app.settings.initSetting(SECTION_NAME, "WIDTH_SIZE", 64));
    var HEIGHT_SIZE = Number(app.settings.initSetting(SECTION_NAME, "HEIGHT_SIZE", 36));

    var SHEET_FOLDER_PATH = app.settings.initSetting(SECTION_NAME, "SHEET_FOLDER_PATH", new File($.fileName).parent.fullName);
    var BOARD_FOLDER_PATH = app.settings.initSetting(SECTION_NAME, "BOARD_FOLDER_PATH", new File($.fileName).parent.fullName);

    var CameraField = { "direction": [{ "name": "FI", "value": "FI", "id": 0, }, { "name": "FO", "value": "FO", "id": 1, }, { "name": "WI", "value": "WI", "id": 2, }, { "name": "WO", "value": "WO", "id": 3, }, { "name": "OL", "value": "OL", "id": 4, }, { "name": "CAM_SHAKE_S", "value": "CAM SHAKE S", "id": 5, }, { "name": "CAM_SHAKE_M", "value": "CAM SHAKE M", "id": 6, }, { "name": "CAM_SHAKE_L", "value": "CAM SHAKE L", "id": 7, }, { "name": "TU", "value": "TU", "id": 8, }, { "name": "TB", "value": "TB", "id": 9, }, { "name": "ZI", "value": "ZI", "id": 10, }, { "name": "ZO", "value": "ZO", "id": 11, }, { "name": "PAN", "value": "PAN", "id": 12, }, { "name": "PAN_UP", "value": "PAN UP", "id": 13, }, { "name": "PAN_DOWN", "value": "PAN DOWN", "id": 14, }, { "name": "TILT", "value": "TILT", "id": 15, }, { "name": "FOLLOW", "value": "FOLLOW", "id": 16, }, { "name": "CU", "value": "CU", "id": 17, }, { "name": "CD", "value": "CD", "id": 18, }, { "name": "DOLLY", "value": "DOLLY", "id": 19, }, { "name": "MULTI", "value": "MULTI", "id": 20, }, { "name": "Fairing", "value": "Fairing", "id": 21, }, { "name": "SL", "value": "SL", "id": 22, }, { "name": "Strobo", "value": "Strobo", "id": 23, }, { "name": "Rotate_TU", "value": "Rotate TU", "id": 24, }, { "name": "Rotate_TB", "value": "Rotate TB", "id": 25, }, { "name": "Handy_S", "value": "Handy S", "id": 26, }, { "name": "Handy_M", "value": "Handy M", "id": 27, }, { "name": "Handy_L", "value": "Handy L", "id": 28, }, { "name": "BL_K", "value": "BL K", "id": 29, }, { "name": "W_K", "value": "W K", "id": 30, }, { "name": "SUBLINA", "value": "SUBLINA", "id": 31, }, { "name": "TFlash", "value": "TFlash", "id": 32, }, { "name": "HI_CON", "value": "HI CON", "id": 33, }, { "name": "Rack_Focus", "value": "Rack Focus", "id": 34, }, { "name": "OverEX", "value": "OverEX", "id": 35, }, { "name": "UnderEX", "value": "UnderEX", "id": 36, }, { "name": "ParsSL", "value": "ParsSL", "id": 37, }, { "name": "JumpSL", "value": "JumpSL", "id": 38, }, { "name": "DF1", "value": "DF1", "id": 39, }, { "name": "DF2", "value": "DF2", "id": 40, }, { "name": "DF3", "value": "DF3", "id": 41, }, { "name": "Fog1", "value": "Fog1", "id": 42, }, { "name": "Fog2", "value": "Fog2", "id": 43, }, { "name": "Fog3", "value": "Fog3", "id": 44, }, { "name": "BOKEH_S", "value": "BOKEH S", "id": 45, }, { "name": "BOKEH_M", "value": "BOKEH M", "id": 46, }, { "name": "BOKEH_L", "value": "BOKEH L", "id": 47, }, { "name": "FIX", "value": "FIX", "id": 48, }, { "name": "PAN_TU", "value": "PAN TU", "id": 49, }, { "name": "PAN_TB", "value": "PAN TB", "id": 50, }, { "name": "FollowPan", "value": "FollowPan", "id": 51, }, { "name": "Rolling", "value": "Rolling", "id": 52, }, { "name": "Q_TU", "value": "Q TU", "id": 53, }, { "name": "Q_TB", "value": "Q TB", "id": 54, }, { "name": "Focus_IN", "value": "Focus IN", "id": 55, }, { "name": "Focus_Out", "value": "Focus Out", "id": 56, }, { "name": "WaveGlass_S", "value": "WaveGlass S", "id": 57, }, { "name": "WaveGlass_M", "value": "WaveGlass M", "id": 58, }, { "name": "WaveGlass_L", "value": "WaveGlass L", "id": 59, }, { "name": "Wipe", "value": "Wipe", "id": 60, }, { "name": "IrisIN", "value": "IrisIN", "id": 61, }, { "name": "IrisOut", "value": "IrisOut", "id": 62, }, { "name": "Insert", "value": "Insert", "id": 63, }, { "name": "CutIN", "value": "CutIN", "id": 64, }, { "name": "Blur1", "value": "Blur1", "id": 65, }, { "name": "Blur2", "value": "Blur2", "id": 66, }, { "name": "Blur3", "value": "Blur3", "id": 67, }, { "name": "WipeIN", "value": "WipeIN", "id": 68, }, { "name": "Bar", "value": "Bar", "id": 69, }, { "name": "Strobo1", "value": "Strobo1", "id": 70, }, { "name": "Strobo2", "value": "Strobo2", "id": 71, }] };

    var data = {
        scriptName: "TimeSheet",
        boardFolderName: "board",
        boardFolderComment: "boardFolder",
        boardItemComment: "boardItem"
    }

    var FieldId = {
        CELL: 0,
        NO_DATA_1: 1,
        NO_DATA_2: 2,
        SERIF: 3,
        NO_DATA_4: 4,
        CAMERA_WORK: 5
    }

    var Instruction = {
        DEFAULT: 0,
    }

    var SerifValue = {
        NAME: 0,
        TEXT: 1
    }

    var Symbol = {
        TICK_1: "SYMBOL_TICK_1",
        TICK_2: "SYMBOL_TICK_2",
        NULL_CELL: "SYMBOL_NULL_CELL",
        HYPHEN: "SYMBOL_HYPHEN"
    }

    var FRAME_RATE = 24;

    var EXTENSION_FILTER = ["Digital TimeSheet File:*.tdts;*.xdts", "Touei Animation Digital TimeSheet File:*.tdts", "Exchange Digital TimeSheet File:*.xdts", "All files:*.*"];


    function buildUI(thisObj) {
        var sheetWindow;
        if (thisObj instanceof Panel) {
            sheetWindow = thisObj;
        } else {
            sheetWindow = new Window("palette", data.scriptName, undefined, { resizeable: true });
        }

        sheetWindow.orientation = "column";
        sheetWindow.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        sheetWindow.spacing = 3;
        sheetWindow.margins = [3, 3, 3, 3];

        var applyPnl = sheetWindow.add("panel");
        applyPnl.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        applyPnl.orientation = "row";
        applyPnl.text = "Apply";

        var sheetCb = applyPnl.add("checkbox");
        sheetCb.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        sheetCb.value = APPLY_SHEET;
        sheetCb.text = "sheet";

        var boardCb = applyPnl.add("checkbox");
        boardCb.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        boardCb.value = ADD_BOARD;
        boardCb.text = "board";

        var durationCb = applyPnl.add("checkbox");
        durationCb.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        durationCb.value = APPLY_DURATION;
        durationCb.text = "duration";

        var sizePnl = sheetWindow.add("panel");
        sizePnl.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        sizePnl.orientation = "row";
        sizePnl.text = "board size (width, height)";
        sizePnl.helpTime = "自動生成するボールドコンポのサイズ";

        var widthEt = sizePnl.add("edittext");
        widthEt.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        widthEt.text = WIDTH_SIZE;
        widthEt.helpTime = "自動生成するボールドコンポのサイズ";

        var heightEt = sizePnl.add("edittext");
        heightEt.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        heightEt.text = HEIGHT_SIZE;
        heightEt.helpTime = "自動生成するボールドコンポのサイズ";

        var sheetPnl = sheetWindow.add("panel");
        sheetPnl.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        sheetPnl.orientation = "row";
        sheetPnl.text = "sheet folder path";
        sheetPnl.helpTime = "自動生成するボールドコンポのサイズ";

        var sheetEt = sheetPnl.add("edittext");
        sheetEt.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        sheetEt.text = new Folder(SHEET_FOLDER_PATH).fsName;

        var folderBtn = sheetPnl.add("button");
        folderBtn.alignment = [ScriptUI.Alignment.RIGHT, ScriptUI.Alignment.TOP];
        folderBtn.text = "Folder";
        folderBtn.onClick = function () {
            var selFolder = new Folder(new Folder(sheetEt.text).fullName).selectDlg("Select sheet folder");
            if (!(selFolder instanceof Folder) || selFolder == null) {
                sheetEt.text = new Folder(SHEET_FOLDER_PATH).fsName;
                return 0;
            }

            sheetEt.text = selFolder.fsName;
        }

        var boardPnl = sheetWindow.add("panel");
        boardPnl.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        boardPnl.orientation = "row";
        boardPnl.text = "board folder path";

        var boardEt = boardPnl.add("edittext");
        boardEt.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        boardEt.text = new Folder(BOARD_FOLDER_PATH).fsName;

        var folderBtn = boardPnl.add("button");
        folderBtn.alignment = [ScriptUI.Alignment.RIGHT, ScriptUI.Alignment.TOP];
        folderBtn.text = "Folder";
        folderBtn.onClick = function () {
            var selFolder = new Folder(new Folder(boardEt.text).fullName).selectDlg("Select board folder");
            if (!(selFolder instanceof Folder) || selFolder == null) {
                boardEt.text = new Folder(BOARD_FOLDER_PATH).fsName;
                return 0;
            }

            boardEt.text = selFolder.fsName;
        }

        var btnGrp = sheetWindow.add("group");
        btnGrp.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];
        btnGrp.orientation = "row";

        var okBtn = btnGrp.add("button", undefined, "OK", { name: "ok" });
        okBtn.alignment = [ScriptUI.Alignment.FILL, ScriptUI.Alignment.TOP];

        okBtn.onClick = function () {
            app.settings.saveSetting(SECTION_NAME, "APPLY_SHEET", String(Number(sheetCb.value)));
            app.settings.saveSetting(SECTION_NAME, "ADD_BOARD", String(Number(boardCb.value)));
            app.settings.saveSetting(SECTION_NAME, "APPLY_DURATION", String(Number(durationCb.value)));

            app.settings.saveSetting(SECTION_NAME, "WIDTH_SIZE", widthEt.text);
            app.settings.saveSetting(SECTION_NAME, "HEIGHT_SIZE", heightEt.text);

            app.settings.saveSetting(SECTION_NAME, "SHEET_FOLDER_PATH", new File(sheetEt.text).fullName);
            app.settings.saveSetting(SECTION_NAME, "BOARD_FOLDER_PATH", new File(boardEt.text).fullName);

            applySheet(sheetCb.value, boardCb.value, durationCb.value, [Number(widthEt.text), Number(heightEt.text)], new Folder(sheetEt.text), new Folder(boardEt.text));
        }

        sheetWindow.layout.layout();
        sheetWindow.onResizing = sheetWindow.onResize = function () {
            this.layout.resize();
        }

        if (!(thisObj instanceof Panel)) {
            sheetWindow.center();
            sheetWindow.show();
        }

        return 0;
    }

    buildUI(this);

    function applySheet(applySheet, addBoard, applyDuration, boardSize, sheetFolder, boardFolder) {
        if (sheetFolder == null || !sheetFolder.exists || sheetFolder == "") {
            sheetFolder = Folder.desktop;
        }

        if (boardFolder == null || !boardFolder.exists || boardFolder == "") {
            boardFolder = Folder.desktop;
        }

        var boardFolderItem = undefined;

        var actComp = app.project.activeItem;
        if (actComp == null || !(actComp instanceof CompItem)) {
            alert("コンポジションをアクティブにして実行してください", data.scriptName);
            return;
        }

        var timeSheetFile = new File(sheetFolder.fullName).openDlg("Select Time Sheet File", EXTENSION_FILTER);
        if (timeSheetFile == null) {
            return 0;
        }

        timeSheetFile.encoding = "utf-8";
        timeSheetFile.open("r");
        var sheetText = timeSheetFile.read();
        timeSheetFile.close();

        var sheetTexts = sheetText.split("\n");

        // header削除
        var headerText = sheetTexts.shift();
        if (headerText != "toeiDigitalTimeSheet Save Data" && headerText != "exchangeDigitalTimeSheet Save Data") {
            alert("デジタルタイムシートファイルを選択してください", data.scriptName);
            return 0;
        }

        var jsonData = JSON.parse(sheetTexts.join(""));
        JSON.stringify(CameraField);

        function getCameraDirectionName(directionId) {
            for (var i = 0; i < CameraField.direction.length; i++) {
                if (CameraField.direction[i].id == directionId) {
                    return CameraField.direction[i].value;
                }
            }

            return "Camera";
        }

        var timeTables = jsonData.timeTables;

        // セルのフィールド取得
        var cellField = null;
        for (var tIdx = 0; tIdx < timeTables.length; tIdx++) {
            for (var hIdx = 0; hIdx < timeTables[tIdx].timeTableHeaders.length; hIdx++) {
                if (timeTables[tIdx].timeTableHeaders[hIdx].fieldId == FieldId.CELL) {
                    cellField = timeTables[tIdx].timeTableHeaders[hIdx];
                }
            }
        }

        var applySheetNumber = -1;

        // シート一覧
        if (timeTables.length > 1) {
            var sheetList = [];
            for (var tIdx = 0; tIdx < timeTables.length; tIdx++) {
                sheetList[sheetList.length] = timeTables[tIdx].name;
            }

            var sheetWin = new Window("dialog", data.scriptName);
            var okBtn = sheetWin.add("statictext", undefined, "適用するシートを選択してください");
            var sheetDdl = sheetWin.add("dropdownlist", undefined, sheetList);
            sheetDdl.selection = 0;

            var okBtn = sheetWin.add("button", undefined, "OK");
            okBtn.onClick = function () {
                sheetWin.close();
                if (sheetDdl.selection == null) {
                    applySheetNumber = -1;
                } else {
                    applySheetNumber = sheetDdl.selection.index;
                }
            }

            sheetWin.center();
            sheetWin.show();
        } else {
            applySheetNumber = 0;
        }

        if (applySheetNumber == -1) {
            return 0;
        }

        var sheetFields = timeTables[applySheetNumber].fields;

        app.beginUndoGroup("タイムシート " + decodeURIComponent(timeSheetFile.name) + " 適用");
        if (applyDuration) {
            actComp.duration = Number(timeTables[applySheetNumber].duration) / FRAME_RATE;
        }

        var skippedLayerNames = [];
        var unusedCellNames = [];

        for (var sIdx = 0; sIdx < sheetFields.length; sIdx++) {
            switch (sheetFields[sIdx].fieldId) {
                case FieldId.CELL:
                    if (!applySheet) {
                        break;
                    }

                    for (var tIdx = 0; tIdx < sheetFields[sIdx].tracks.length; tIdx++) {
                        // ここで一つのセルになる
                        var track = sheetFields[sIdx].tracks[tIdx];
                        var blindsValues = [];
                        var trackSeconds = [];
                        var trackValues = [];
                        var trackInstructionValues = [];

                        // 時間データとセル番号を取得する
                        for (var fIdx = 0; fIdx < track.frames.length; fIdx++) {
                            for (var dIdx = 0; dIdx < track.frames[fIdx].data.length; dIdx++) {
                                if (Number(track.frames[fIdx].data[dIdx].id) != Instruction.DEFAULT) {
                                    continue;
                                }

                                if (track.frames[fIdx].data[dIdx].values == null) {
                                    continue;
                                }

                                for (var vIdx = 0; vIdx < track.frames[fIdx].data[dIdx].values.length; vIdx++) {
                                    // 中割り・逆シートなら追加しない
                                    if (track.frames[fIdx].data[dIdx].values[vIdx] == Symbol.TICK_1 || track.frames[fIdx].data[dIdx].values[vIdx] == Symbol.TICK_2) {
                                        continue;
                                    }

                                    // 指示文字列の場合
                                    if (isNaN(Number(track.frames[fIdx].data[dIdx].values[vIdx]))) {
                                        trackInstructionValues[trackInstructionValues.length] = track.frames[fIdx].data[dIdx].values[vIdx];
                                        blindsValues[blindsValues.length] = 100;

                                        if (vIdx == 0) {
                                            trackSeconds[trackSeconds.length] = Number(track.frames[fIdx].frame) / FRAME_RATE;
                                            trackValues[trackValues.length] = 0;
                                        } else {
                                            if (track.frames[fIdx].data[dIdx].values[vIdx] == Symbol.NULL_CELL) {
                                                trackSeconds[trackSeconds.length] = trackSeconds[trackSeconds.length - 1];
                                                trackValues[trackValues.length] = trackValues[trackValues.length - 1];
                                            } else {
                                                trackSeconds[trackSeconds.length] = Number(track.frames[fIdx].frame) / FRAME_RATE;
                                                trackValues[trackValues.length] = (track.frames[fIdx].data[dIdx].values[vIdx] / FRAME_RATE) - actComp.frameDuration;
                                            }
                                        }
                                    } else {
                                        trackInstructionValues[trackInstructionValues.length] = "";
                                        blindsValues[blindsValues.length] = 0;

                                        // 空セルなら一つ前のフレームにキー
                                        trackSeconds[trackSeconds.length] = Number(track.frames[fIdx].frame) / FRAME_RATE;
                                        trackValues[trackValues.length] = (track.frames[fIdx].data[dIdx].values[vIdx] / FRAME_RATE) - actComp.frameDuration;
                                    }
                                }
                            }
                        }

                        var targetLayers = actComp.selectedLayers;
                        if (targetLayers.length == 0) {
                            for (var i = 1; i <= actComp.layers.length; i++) {
                                targetLayers[targetLayers.length] = actComp.layer(i);
                            }
                        }

                        var sheetFrameMax = Number.MIN_VALUE;
                        for (var f = 0; f <= trackValues.length; f++) {
                            if (sheetFrameMax < trackValues[f] * FRAME_RATE) {
                                sheetFrameMax = trackValues[f] * FRAME_RATE;
                            }
                        }


                        for (var i = 0; i < targetLayers.length; i++) {
                            if (!targetLayers[i].canSetTimeRemapEnabled) {
                                continue;
                            }

                            if (targetLayers[i].name.toLowerCase() != cellField.names[track.trackNo].toLowerCase()) {
                                continue;
                            }

                            var sequenceFrameMax = targetLayers[i].source.duration / targetLayers[i].source.frameDuration;
                            if (sequenceFrameMax < sheetFrameMax) {
                                skippedLayerNames[skippedLayerNames.length] = targetLayers[i].index + ": " + targetLayers[i].name + " (セルの数 " + sequenceFrameMax + ", シート最大値 " + String(Number(sheetFrameMax) + 1) + ")";
                                continue;
                            }

                            if (sequenceFrameMax != Number(sheetFrameMax + 1)) {
                                unusedCellNames[unusedCellNames.length] = targetLayers[i].index + ": " + targetLayers[i].name + " (セルの数 " + sequenceFrameMax + ", シート最大値 " + String(Number(sheetFrameMax) + 1) + ")";
                            }

                            targetLayers[i].timeRemapEnabled = true;

                            while (targetLayers[i].property("ADBE Time Remapping").numKeys > 1) {
                                targetLayers[i].property("ADBE Time Remapping").removeKey(targetLayers[i].property("ADBE Time Remapping").numKeys);
                            }

                            targetLayers[i].property("ADBE Time Remapping").setValuesAtTimes(trackSeconds, trackValues);
                            for (var j = 1; j <= targetLayers[i].property("ADBE Time Remapping").numKeys; j++) {
                                targetLayers[i].property("ADBE Time Remapping").setInterpolationTypeAtKey(j, KeyframeInterpolationType.HOLD, KeyframeInterpolationType.HOLD);
                            }

                            while (targetLayers[i].property("ADBE Effect Parade").property("ADBE Venetian Blinds") != null) {
                                targetLayers[i].property("ADBE Effect Parade").property("ADBE Venetian Blinds").remove();
                            }

                            var blindFx = targetLayers[i].property("ADBE Effect Parade").addProperty("ADBE Venetian Blinds");
                            blindFx.property("ADBE Venetian Blinds-0001").setValuesAtTimes(trackSeconds, blindsValues);

                            // 先頭が空セルの場合の処理(時間データはタイムリマップの配列と共通のため、ここで処理する)
                            // 先頭のキーフレームが0にない・先頭のブラインド値が0
                            if (blindFx.property("ADBE Venetian Blinds-0001").keyTime(1) != 0 && blindFx.property("ADBE Venetian Blinds-0001").keyValue(1) == 0) {
                                blindFx.property("ADBE Venetian Blinds-0001").setValueAtTime(0, 100);
                            }


                            for (var j = 1; j <= blindFx.property("ADBE Venetian Blinds-0001").numKeys; j++) {
                                blindFx.property("ADBE Venetian Blinds-0001").setInterpolationTypeAtKey(j, KeyframeInterpolationType.HOLD, KeyframeInterpolationType.HOLD);
                            }
                        }

                        // ダブラシなど対応 (一度シートを全て確認後,タイムリマップが未反映のものに対して走査)
                        for (var i = 0; i < targetLayers.length; i++) {
                            if (!targetLayers[i].canSetTimeRemapEnabled) {
                                continue;
                            }

                            if (targetLayers[i].timeRemapEnabled) {
                                continue;
                            }

                            var tempName = targetLayers[i].name.split("_");
                            // アンダーバー区切りで先頭文字列が一致する時,同セルとみなす
                            // a_wxpなどはa_wxpのシートがなければaのシートが適用される
                            // a_shita_wxpなどはa_shita_wxpがなければa_shita,aの順に適用するシートが決定される
                            while (tempName.pop()) {
                                if (tempName.join("_").toLowerCase() != cellField.names[track.trackNo].toLowerCase()) {
                                    continue;
                                }

                                var sequenceFrameMax = targetLayers[i].source.duration / targetLayers[i].source.frameDuration;
                                if (sequenceFrameMax < sheetFrameMax) {
                                    skippedLayerNames[skippedLayerNames.length] = targetLayers[i].index + ": " + targetLayers[i].name + " (セルの数 " + sequenceFrameMax + ", シート最大値 " + String(Number(sheetFrameMax) + 1) + ")";
                                    continue;
                                }

                                if (sequenceFrameMax != sheetFrameMax) {
                                    unusedCellNames[unusedCellNames.length] = targetLayers[i].index + ": " + targetLayers[i].name + " (セルの数 " + sequenceFrameMax + ", シート最大値 " + String(Number(sheetFrameMax) + 1) + ")";
                                }

                                targetLayers[i].timeRemapEnabled = true;

                                while (targetLayers[i].property("ADBE Time Remapping").numKeys > 1) {
                                    targetLayers[i].property("ADBE Time Remapping").removeKey(targetLayers[i].property("ADBE Time Remapping").numKeys);
                                }

                                targetLayers[i].property("ADBE Time Remapping").setValuesAtTimes(trackSeconds, trackValues);
                                for (var j = 1; j <= targetLayers[i].property("ADBE Time Remapping").numKeys; j++) {
                                    targetLayers[i].property("ADBE Time Remapping").setInterpolationTypeAtKey(j, KeyframeInterpolationType.HOLD, KeyframeInterpolationType.HOLD);
                                }

                                while (targetLayers[i].property("ADBE Effect Parade").property("ADBE Venetian Blinds") != null) {
                                    targetLayers[i].property("ADBE Effect Parade").property("ADBE Venetian Blinds").remove();
                                }

                                var blindFx = targetLayers[i].property("ADBE Effect Parade").addProperty("ADBE Venetian Blinds");
                                blindFx.property("ADBE Venetian Blinds-0001").setValuesAtTimes(trackSeconds, blindsValues);

                                // 先頭が空セルの場合の処理(時間データはタイムリマップの配列と共通のため、ここで処理する)
                                // 先頭のキーフレームが0にない・先頭のブラインド値が0
                                if (blindFx.property("ADBE Venetian Blinds-0001").keyTime(1) != 0 && blindFx.property("ADBE Venetian Blinds-0001").keyValue(1) == 0) {
                                    blindFx.property("ADBE Venetian Blinds-0001").setValueAtTime(0, 100);
                                }

                                for (var j = 1; j <= blindFx.property("ADBE Venetian Blinds-0001").numKeys; j++) {
                                    blindFx.property("ADBE Venetian Blinds-0001").setInterpolationTypeAtKey(j, KeyframeInterpolationType.HOLD, KeyframeInterpolationType.HOLD);
                                }
                            }
                        }
                    }
                    break;

                case FieldId.SERIF:
                case FieldId.CAMERA_WORK:
                    if (!addBoard) {
                        break;
                    }

                    for (var tIdx = 0; tIdx < sheetFields[sIdx].tracks.length; tIdx++) {
                        // ここで一つのトラックになる
                        var track = sheetFields[sIdx].tracks[tIdx];
                        var boardSerifSeconds = [];
                        var boardNameValues = [];
                        var boardSerifValues = [];
                        var boardSerifDurations = [];

                        // 時間データとセル番号を取得する
                        for (var fIdx = 0; fIdx < track.frames.length; fIdx++) {
                            for (var dIdx = 0; dIdx < track.frames[fIdx].data.length; dIdx++) {
                                if (Number(track.frames[fIdx].data[dIdx].id) != Instruction.DEFAULT) {
                                    continue;
                                }

                                if (track.frames[fIdx].data[dIdx].values == null) {
                                    continue;
                                }

                                if (track.frames[fIdx].data[dIdx].values[SerifValue.NAME] != Symbol.HYPHEN) {
                                    boardSerifSeconds[boardSerifSeconds.length] = Number(track.frames[fIdx].frame) / FRAME_RATE;
                                    if (sheetFields[sIdx].fieldId == FieldId.SERIF) {
                                        boardNameValues[boardNameValues.length] = track.frames[fIdx].data[dIdx].values[SerifValue.NAME];
                                    } else {
                                        boardNameValues[boardNameValues.length] = getCameraDirectionName(Number(track.frames[fIdx].data[dIdx].values[SerifValue.NAME]));
                                    }
                                    boardSerifDurations[boardSerifDurations.length] = Number(track.frames[track.frames.length - 1].frame);
                                    if (track.frames[fIdx].data[dIdx].values.length <= 1) {
                                        boardSerifValues[boardSerifValues.length] = "";
                                    } else {
                                        boardSerifValues[boardSerifValues.length] = track.frames[fIdx].data[dIdx].values[SerifValue.TEXT];
                                    }
                                }

                                // 現在のトラックの一つ前のボールドのデュレーションを計算する
                                boardSerifDurations[boardSerifDurations.length - 1] = (Number(track.frames[fIdx].frame) + 1) / FRAME_RATE;
                            }
                        }

                        for (var itemIndex = 1; itemIndex <= app.project.numItems; itemIndex++) {
                            if (app.project.item(itemIndex).comment == data.boardFolderComment) {
                                boardFolderItem = app.project.item(itemIndex);
                                break;
                            }
                        }

                        if (boardFolderItem == undefined) {
                            boardFolderItem = app.project.items.addFolder(data.boardFolderComment);
                            boardFolderItem.comment = data.boardFolderComment;
                        }

                        for (var i = 0; i < boardSerifSeconds.length; i++) {
                            var boardImageFile = new Folder(boardFolder.fullName).getBoardImage(boardNameValues[i]);

                            var boardItem = undefined;
                            for (var itemIndex = 1; itemIndex <= app.project.numItems; itemIndex++) {
                                if (boardImageFile != null) {
                                    if (app.project.item(itemIndex) instanceof FootageItem) {
                                        if (app.project.item(itemIndex).file == boardImageFile) {
                                            boardItem = app.project.item(itemIndex);
                                            break;
                                        }
                                    }
                                }

                                if (app.project.item(itemIndex).comment == boardNameValues[i]) {
                                    boardItem = app.project.item(itemIndex);
                                    break;
                                }
                            }

                            if (boardImageFile == null) {
                                if (boardItem == undefined) {
                                    boardItem = app.project.items.generateBoardComp(boardNameValues[i], boardSize[0], boardSize[1], actComp, [0, 0, 0], [1, 1, 1]);
                                }
                            } else {
                                if (boardItem == undefined) {
                                    boardItem = app.project.importFile(new ImportOptions(boardImageFile));
                                }
                            }

                            boardItem.comment = boardNameValues[i];
                            boardItem.parentFolder = boardFolderItem;

                            var boardLayer = actComp.layers.add(boardItem);
                            boardLayer.inPoint = boardSerifSeconds[i];
                            boardLayer.outPoint = boardSerifDurations[i];
                            if (sheetFields[sIdx].fieldId == FieldId.SERIF) {
                                boardLayer.property("ADBE Transform Group").property("ADBE Position").setValue(
                                    [actComp.width * 0.9 - boardLayer.width / 2, actComp.height * 0.1 + boardLayer.height / 2 + boardLayer.height * tIdx + (boardLayer.height / 10 * tIdx)]
                                );
                            } else {
                                boardLayer.property("ADBE Transform Group").property("ADBE Position").setValue(
                                    [actComp.width * 0.8 - boardLayer.width / 2, actComp.height * 0.1 + boardLayer.height / 2 + boardLayer.height * tIdx + (boardLayer.height / 10 * tIdx)]
                                );
                            }

                            if (boardSerifValues[i] != "") {
                                var curMarker = new MarkerValue(boardSerifValues[i]);
                                boardLayer.property("ADBE Marker").setValueAtTime(boardSerifSeconds[i], curMarker);
                            }
                        }
                    }
                    break;

                default:
                    break;
            }
        }

        var errMsg = "";
        if (skippedLayerNames.length != 0) {
            errMsg += "以下のレイヤーは素材の尺がシートの最大値に足りないため,タイムリマップが適用されませんでした.\n\n" + skippedLayerNames.join("\n");
        }

        if (unusedCellNames.length != 0) {
            if (skippedLayerNames.length != 0) {
                errMsg += "\n\n\n";
            }
            errMsg += "以下のレイヤーは素材の尺がシートの最大値より大きいため,使用されていないフレームがあります.\n\n" + unusedCellNames.join("\n");
        }

        if (skippedLayerNames.length != 0 || unusedCellNames.length != 0) {
            alert(errMsg, data.scriptName);
        }

        app.endUndoGroup();
        return 0;
    }

    return 0;
}).call(this);
